from kybra import blob, nat32, nat64, opt, Record, Variant

BitcoinAddress = str
BlockHash = blob


class BitcoinNetwork(Variant, total=False):
    Mainnet: None
    Regtest: None
    Testnet: None


class GetBalanceArgs(Record):
    address: BitcoinAddress
    min_confirmations: opt[nat32]
    network: BitcoinNetwork


class GetCurrentFeePercentilesArgs(Record):
    network: BitcoinNetwork


Page = blob


class UtxosFilter(Variant, total=False):
    MinConfirmations: nat32
    Page: Page


class GetUtxosArgs(Record):
    address: BitcoinAddress
    filter: opt[UtxosFilter]
    network: BitcoinNetwork


class Outpoint(Record):
    txid: blob
    vout: nat32


Satoshi = nat64


class Utxo(Record):
    height: nat32
    outpoint: Outpoint
    value: Satoshi


class GetUtxosResult(Record):
    next_page: opt[Page]
    tip_block_hash: BlockHash
    tip_height: nat32
    utxos: list[Utxo]


MillisatoshiPerByte = nat64


class SendTransactionArgs(Record):
    transaction: blob
    network: BitcoinNetwork


class SendTransactionError(Variant, total=False):
    MalformedTransaction: None
    QueueFull: None



# Copyright 2021 DFINITY Stiftung

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

# The original license for the DFINITY code can be found here: https://github.com/dfinity/ic/blob/master/LICENSE
# This file contains derivative works licensed as MIT: https://github.com/demergent-labs/azle/blob/main/LICENSE

# Taken in part from: https://github.com/dfinity/interface-spec/blob/master/spec/ic.did

from kybra import blob, Canister, method, Principal
from kybra.canisters.management.basic import CanisterStatusResult, DepositCyclesArgs, DeleteCanisterArgs, CreateCanisterArgs, CreateCanisterResult, InstallCodeArgs, UninstallCodeArgs, UpdateSettingsArgs, StartCanisterArgs, StopCanisterArgs, CanisterStatusArgs, ProvisionalCreateCanisterWithCyclesArgs, ProvisionalCreateCanisterWithCyclesResult, ProvisionalTopUpCanisterArgs
from kybra.canisters.management.http import HttpRequestArgs, HttpResponse
from kybra.canisters.management.bitcoin import GetBalanceArgs, GetCurrentFeePercentilesArgs, GetUtxosArgs, GetUtxosResult, MillisatoshiPerByte, Satoshi, SendTransactionArgs

# TODO change the return types to void


class ManagementCanister(Canister):
    @method
    def bitcoin_get_balance(self, args: GetBalanceArgs) -> Satoshi: ...

    @method
    def bitcoin_get_current_fee_percentiles(
        self, args: GetCurrentFeePercentilesArgs) -> list[MillisatoshiPerByte]: ...

    @method
    def bitcoin_get_utxos(self, args: GetUtxosArgs) -> GetUtxosResult: ...

    @method
    def bitcoin_send_transaction(self, args: SendTransactionArgs) -> None: ...

    @method
    def create_canister(
        self, args: CreateCanisterArgs) -> CreateCanisterResult: ...

    @method
    def update_settings(self, args: UpdateSettingsArgs) -> None: ...

    @method
    def install_code(self, args: InstallCodeArgs) -> None: ...

    @method
    def uninstall_code(self, args: UninstallCodeArgs) -> None: ...

    @method
    def start_canister(self, args: StartCanisterArgs) -> None: ...

    @method
    def stop_canister(self, args: StopCanisterArgs) -> None: ...

    @method
    def canister_status(
        self, args: CanisterStatusArgs) -> CanisterStatusResult: ...

    @method
    def delete_canister(self, args: DeleteCanisterArgs) -> None: ...

    @method
    def deposit_cycles(self, args: DepositCyclesArgs) -> None: ...

    @method
    def raw_rand(self) -> blob: ...

    @method
    def provisional_create_canister_with_cycles(
        self, args: ProvisionalCreateCanisterWithCyclesArgs) -> ProvisionalCreateCanisterWithCyclesResult: ...

    @method
    def provisional_top_up_canister(
        self, args: ProvisionalTopUpCanisterArgs) -> None: ...

    @method
    def http_request(self, args: HttpRequestArgs) -> HttpResponse: ...


management_canister = ManagementCanister(Principal.from_str('aaaaa-aa'))



from kybra import blob, Func, nat, nat64, opt, Query, Record, Variant
from typing import TypeAlias


class HttpMethod(Variant, total=False):
    get: None
    head: None
    post: None


class HttpHeader(Record):
    name: str
    value: str


class HttpResponse(Record):
    status: nat
    headers: list[HttpHeader]
    body: blob


class HttpTransformArgs(Record):
    response: HttpResponse
    context: blob


HttpTransformFunc: TypeAlias = Func(Query[[HttpTransformArgs], HttpResponse])


class HttpTransform(Record):
    function: HttpTransformFunc
    context: blob


class HttpRequestArgs(Record):
    url: str
    max_response_bytes: opt[nat64]
    method: HttpMethod
    headers: list[HttpHeader]
    body: opt[blob]
    transform: opt[HttpTransform]



import modulegraph.modulegraph
import os
from pathlib import Path
import re
import shutil
import subprocess
import sys
import time
from typing import Any, Callable

import kybra
from kybra.colors import red, yellow, green, dim
from kybra.timed import timed, timed_inline
from kybra.types import Args, Paths


@timed
def main():
    args = parse_args_or_exit(sys.argv)
    paths = create_paths(args)
    is_verbose = args["flags"]["verbose"]
    is_initial_compile = detect_initial_compile(paths["gzipped_wasm"])

    # This is the name of the canister passed into python -m kybra from the dfx.json build command
    canister_name = args["canister_name"]

    verbose_mode_qualifier = " in verbose mode" if is_verbose else ""

    print(f"\nBuilding canister {green(canister_name)}{verbose_mode_qualifier}\n")

    if is_initial_compile:
        print(
            yellow(
                "Initial build takes a few minutes. Don't panic. Subsequent builds will be faster.\n"
            )
        )

    # Copy all of the Rust project structure from the pip package to an area designed for Rust compiling
    shutil.copytree(paths["compiler"], paths["canister"], dirs_exist_ok=True)

    # Add CARGO_TARGET_DIR to env for all cargo commands
    cargo_env = {**os.environ.copy(), "CARGO_TARGET_DIR": paths["target"]}

    compile_python_or_exit(
        paths, cargo_env, verbose=is_verbose, label="[1/3] 🔨 Compiling Python..."
    )

    generate_candid_file_or_exit(
        paths,
        cargo_env,
        verbose=is_verbose,
        label=f"[2/3] 📝 Generating Candid file...{encourage_patience(is_initial_compile)}",
    )

    build_wasm_binary_or_exit(
        paths,
        cargo_env,
        verbose=is_verbose,
        label=f"[3/3] 🚧 Building Wasm binary...{show_empathy(is_initial_compile)}",
    )

    print(f"\n🎉 Built canister {green(canister_name)} at {dim(paths['gzipped_wasm'])}")


def parse_args_or_exit(args: list[str]) -> Args:
    args = args[1:]  # Discard the path to kybra

    flags = [arg for arg in args if (arg.startswith("-") or arg.startswith("--"))]
    args = [arg for arg in args if not (arg.startswith("-") or arg.startswith("--"))]

    if len(args) == 0:
        print(f"\nkybra {kybra.__version__}")
        print("\nUsage: kybra [-v|--verbose] <canister_name> <entry_point> <did_path>")
        sys.exit(0)

    if len(args) != 3:
        print(red("\n💣 wrong number of arguments\n"))
        print("Usage: kybra [-v|--verbose] <canister_name> <entry_point> <did_path>")
        print("\n💀 Build failed!")
        sys.exit(1)

    return {
        "empty": False,
        "flags": {"verbose": "--verbose" in flags or "-v" in flags},
        "canister_name": args[0],
        "entry_point": args[1],
        "did_path": args[2],
    }


def create_paths(args: Args) -> Paths:
    canister_name = args["canister_name"]

    # This is the path to the developer's entry point Python file passed into python -m kybra from the dfx.json build command
    py_entry_file_path = args["entry_point"]

    # This is the Python module name of the developer's Python project, derived from the entry point Python file passed into python -m kybra from the dfx.json build command
    py_entry_module_name = Path(py_entry_file_path).stem

    # This is the location of all code used to generate the final canister Rust code
    canister_path = f".dfx/kybra/{canister_name}"

    # We want to bundle/gather all Python files into the python_source directory for RustPython freezing
    # The location that Kybra will look to when running py_freeze!
    # py_freeze! will compile all of the Python code in the directory recursively (modules must have an __init__.py to be included)
    python_source_path = f"{canister_path}/python_source"

    py_file_names_file_path = f"{canister_path}/file_names.txt"

    # This is the path to the developer's Candid file passed into python -m kybra from the dfx.json build command
    did_path = args["did_path"]

    # This is the path to the Kybra compiler Rust code delivered with the Python package
    compiler_path = os.path.dirname(kybra.__file__) + "/compiler"

    # This is the final generated Rust file that is the canister
    lib_path = f"{canister_path}/src/lib.rs"

    # This is the location of the Candid file generated from the final generated Rust file
    generated_did_path = f"{canister_path}/index.did"

    # This is the Rust target directory
    target_path = f"{canister_path}/target"

    # This is the unzipped generated Wasm that is the canister
    wasm_path = f"{target_path}/wasm32-unknown-unknown/release/{canister_name}.wasm"

    # This is the final zipped generated Wasm that will actually run on the Internet Computer
    gzipped_wasm_path = f"{wasm_path}.gz"

    # This is where we store custom Python modules, such as stripped-down versions of stdlib modules
    custom_modules_path = f"{compiler_path}/custom_modules"

    return {
        "py_entry_file": py_entry_file_path,
        "py_entry_module_name": py_entry_module_name,
        "canister": canister_path,
        "python_source": python_source_path,
        "py_file_names_file": py_file_names_file_path,
        "did": did_path,
        "compiler": compiler_path,
        "lib": lib_path,
        "generated_did": generated_did_path,
        "target": target_path,
        "wasm": wasm_path,
        "gzipped_wasm": gzipped_wasm_path,
        "custom_modules": custom_modules_path,
    }


def detect_initial_compile(gzipped_wasm_path: str) -> bool:
    return not os.path.exists(gzipped_wasm_path)


@timed_inline
def compile_python_or_exit(
    paths: Paths, cargo_env: dict[str, str], verbose: bool = False
):
    bundle_python_code(paths)
    add_wasm_compilation_target_or_exit(verbose)
    install_ic_cdk_optimizer_or_exit(verbose)
    run_kybra_generate_or_exit(paths, cargo_env, verbose)
    run_rustfmt_or_exit(paths, verbose)


def encourage_patience(is_initial_compile: bool) -> str:
    return " (be patient, this will take a while)" if is_initial_compile else ""


def bundle_python_code(paths: Paths):
    # Begin module bundling/gathering process
    path = list(filter(lambda x: x.startswith(os.getcwd()), sys.path)) + [
        os.path.dirname(paths["py_entry_file"])
    ]

    graph = modulegraph.modulegraph.ModuleGraph(path)
    entry_point = graph.run_script(paths["py_entry_file"])  # type: ignore

    python_source_path = paths["python_source"]

    if os.path.exists(python_source_path):
        shutil.rmtree(python_source_path)

    os.makedirs(python_source_path)

    # Copy our custom Python modules into the python_source directory
    shutil.copytree(paths["custom_modules"], python_source_path, dirs_exist_ok=True)

    flattened_graph = list(graph.flatten(start=entry_point))  # type: ignore

    for node in flattened_graph:  # type: ignore
        if type(node) == modulegraph.modulegraph.Script:  # type: ignore
            shutil.copy(
                node.filename, f"{python_source_path}/{os.path.basename(node.filename)}"  # type: ignore
            )

        if type(node) == modulegraph.modulegraph.SourceModule:  # type: ignore
            shutil.copy(
                node.filename, f"{python_source_path}/{os.path.basename(node.filename)}"  # type: ignore
            )

        if type(node) == modulegraph.modulegraph.Package:  # type: ignore
            shutil.copytree(
                node.packagepath[0],  # type: ignore
                f"{python_source_path}/{node.identifier}",  # type: ignore
                dirs_exist_ok=True,
            )

        if type(node) == modulegraph.modulegraph.NamespacePackage:  # type: ignore
            shutil.copytree(
                node.packagepath[0],  # type: ignore
                f"{python_source_path}/{node.identifier}",  # type: ignore
                dirs_exist_ok=True,
            )

    py_file_names = list(  # type: ignore
        filter(
            lambda filename: filename is not None,  # type: ignore
            map(
                lambda node: node.filename,  # type: ignore
                filter(
                    lambda node: node.filename  # type: ignore
                    is not "-",  # This filters out namespace packages
                    flattened_graph,  # type: ignore
                ),  # type: ignore
            ),  # type: ignore
        )  # type: ignore
    )

    create_file(paths["py_file_names_file"], ",".join(py_file_names))  # type: ignore


def add_wasm_compilation_target_or_exit(verbose: bool = False):
    add_wasm_target_result = subprocess.run(
        ["rustup", "target", "add", "wasm32-unknown-unknown"],
        capture_output=not verbose,
    )

    if add_wasm_target_result.returncode != 0:
        print(red("\n💣 Unable to add wasm32-unknown-unknown compilation target\n"))
        print(add_wasm_target_result.stderr.decode("utf-8"))
        print("💀 Build failed")
        sys.exit(1)


# TODO reconcile this with Dan's code
def install_ic_cdk_optimizer_or_exit(verbose: bool = False):
    # TODO this should eventually be replaced with ic-wasm once this is resolved: https://forum.dfinity.org/t/wasm-module-contains-a-function-that-is-too-complex/15407/43?u=lastmjs
    subprocess.run(
        ["cargo", "install", "ic-cdk-optimizer", "--version=0.3.4"],
        capture_output=not verbose,
    )
    subprocess.run(
        ["cargo", "install", "ic-wasm", "--version=0.3.0"], capture_output=not verbose
    )

    # if install_cdk_optimizer_result.returncode != 0:
    #     print(red('\n💣 Unable to install dependency "ic-cdk-optimizer"\n'))
    #     print(install_cdk_optimizer_result.stderr.decode("utf-8"))
    #     print("💀 Build failed")
    #     sys.exit(1)


def run_kybra_generate_or_exit(paths: Paths, cargo_env: dict[str, str], verbose: bool):
    # Generate the Rust code
    kybra_generate_result = subprocess.run(
        [
            "cargo",
            "run",
            f"--manifest-path={paths['canister']}/kybra_generate/Cargo.toml",
            paths["py_file_names_file"],
            paths["py_entry_module_name"],
            paths["lib"],
        ],
        capture_output=not verbose,
        env=cargo_env,
    )

    if kybra_generate_result.returncode != 0:
        print(
            red("\n💣 Something about your Python code violates Kybra's requirements\n")
        )
        print(parse_kybra_generate_error(kybra_generate_result.stderr))
        print(
            "\nIf you are unable to decipher the error above, reach out in the #typescript"
        )
        print("channel of the DFINITY DEV OFFICIAL discord:")
        print("\nhttps://discord.com/channels/748416164832608337/1019372359775440988\n")
        print("💀 Build failed")
        sys.exit(1)


def parse_kybra_generate_error(stdout: bytes) -> str:
    err = stdout.decode("utf-8")
    std_err_lines = err.splitlines()
    try:
        line_where_error_message_starts = next(
            i
            for i, v in enumerate(std_err_lines)
            if v.startswith("thread 'main' panicked at '")
        )
        line_where_error_message_ends = next(
            i for i, v in enumerate(std_err_lines) if "', src/" in v
        )
    except:
        return (
            "The underlying cause is likely at the bottom of the following output:\n\n"
            + err
        )

    err_lines = std_err_lines[
        line_where_error_message_starts : line_where_error_message_ends + 1
    ]
    err_lines[0] = err_lines[0].replace("thread 'main' panicked at '", "")
    err_lines[-1] = re.sub("', src/.*", "", err_lines[-1])

    return red("\n".join(err_lines))


def run_rustfmt_or_exit(paths: Paths, verbose: bool = False):
    rustfmt_result = subprocess.run(
        ["rustfmt", "--edition=2018", paths["lib"]], capture_output=not verbose
    )

    if rustfmt_result.returncode != 0:
        print(
            red(
                "\n💣 Kybra has experienced an internal error while trying to\n   format your generated rust canister"
            )
        )
        print(
            f'\nPlease open an issue at https://github.com/demergent-labs/kybra/issues/new\nincluding this message and the following error:\n\n {red(rustfmt_result.stderr.decode("utf-8"))}'
        )
        print("💀 Build failed")
        sys.exit(1)


@timed_inline
def build_wasm_binary_or_exit(
    paths: Paths, cargo_env: dict[str, str], verbose: bool = False
):
    # Compile the generated Rust code
    cargo_build_result = subprocess.run(
        [
            "cargo",
            "build",
            f"--manifest-path={paths['canister']}/Cargo.toml",
            "--target=wasm32-unknown-unknown",
            "--package=kybra_generated_canister",
            "--release",
        ],
        capture_output=not verbose,
        env=cargo_env,
    )

    if cargo_build_result.returncode != 0:
        print(red("\n💣 Error building Wasm binary:"))
        print(cargo_build_result.stderr.decode("utf-8"))
        print("💀 Build failed")
        sys.exit(1)

    cargo_bin_root = (
        os.environ.get("CARGO_INSTALL_ROOT")
        or os.environ.get("CARGO_HOME")
        or f'{os.environ["HOME"]}/.cargo'
    )

    # Optimize the Wasm binary
    # TODO this should eventually be replaced with ic-wasm once this is resolved: https://forum.dfinity.org/t/wasm-module-contains-a-function-that-is-too-complex/15407/43?u=lastmjs
    optimization_result = subprocess.run(
        [
            f"{cargo_bin_root}/bin/ic-cdk-optimizer",
            f"{paths['target']}/wasm32-unknown-unknown/release/kybra_generated_canister.wasm",
            f"-o={paths['wasm']}",
        ],
        capture_output=not verbose,
    )
    # optimization_result = subprocess.run(
    #     [
    #         f"{cargo_bin_root}/bin/ic-wasm",
    #         f"{paths['target']}/wasm32-unknown-unknown/release/kybra_generated_canister.wasm",
    #         f"-o={paths['wasm']}",
    #         "shrink",
    #     ],
    #     capture_output=not verbose,
    # )

    if optimization_result.returncode != 0:
        print(red("\n💣 Error optimizing generated Wasm:"))
        print(optimization_result.stderr.decode("utf-8"))
        print("💀 Build failed")
        sys.exit(1)

    add_metadata_to_wasm_or_exit(paths, verbose=verbose)

    # gzip the Wasm binary
    os.system(f"gzip -f -k {paths['wasm']}")


def add_metadata_to_wasm_or_exit(paths: Paths, verbose: bool = False):
    add_candid_to_wasm_result = subprocess.run(
        [
            "ic-wasm",
            paths["wasm"],
            "-o",
            paths["wasm"],
            "metadata",
            "candid:service",
            "-f",
            paths["did"],
            "-v",
            "public",
        ],
        capture_output=not verbose,
    )

    if add_candid_to_wasm_result.returncode != 0:
        print(red("\n💣 Error adding candid to Wasm:"))
        print(add_candid_to_wasm_result.stderr.decode("utf-8"))
        print("💀 Build failed")
        sys.exit(1)

    add_cdk_info_to_wasm_result = subprocess.run(
        [
            "ic-wasm",
            paths["wasm"],
            "-o",
            paths["wasm"],
            "metadata",
            "cdk",
            "-d",
            f"kybra {kybra.__version__}",
            "-v",
            "public",
        ],
        capture_output=not verbose,
    )

    if add_cdk_info_to_wasm_result.returncode != 0:
        print(red("\n💣 Error adding cdk name/version to Wasm:"))
        print(add_cdk_info_to_wasm_result.stderr.decode("utf-8"))
        print("💀 Build failed")
        sys.exit(1)


def show_empathy(is_initial_compile: bool) -> str:
    return (
        " (❤ hang in there, this will be faster next time)"
        if is_initial_compile
        else ""
    )


@timed_inline
def generate_candid_file_or_exit(
    paths: Paths, cargo_env: dict[str, str], verbose: bool = False
):
    generate_candid_result = subprocess.run(
        [
            "cargo",
            "test",
            f"--manifest-path={paths['canister']}/Cargo.toml",
        ],
        capture_output=not verbose,
        env=cargo_env,
    )

    if generate_candid_result.returncode != 0:
        print(red("\n💣 Error generating candid:"))
        print(generate_candid_result.stderr.decode("utf-8"))
        print("💀 Build failed")
        sys.exit(1)

    # Copy the generated Candid file to the developer's source directory
    os.system(f"cp {paths['generated_did']} {paths['did']}")


def create_file(file_path: str, contents: str):
    file = open(file_path, "w")
    file.write(contents)
    file.close()


def inline_timed(
    label: str,
    body: Callable[..., Any],
    *args: Any,
    verbose: bool = False,
    **kwargs: Any,
) -> float:
    print(label)
    start_time = time.time()
    body(*args, verbose=verbose, **kwargs)
    end_time = time.time()
    duration = end_time - start_time

    if verbose:
        print(f"{label} finished in {round(duration, 2)}s")
    else:
        move_cursor_up_one_line = "\x1b[1A"
        print(f'{move_cursor_up_one_line}{label} {dim(f"{round(duration, 2)}s")}')

    return end_time - start_time


main()



import time
from typing import Any, Callable, TypeVar

from kybra.colors import dim

T = TypeVar('T')

def timed(func: Callable[..., T]) -> Callable[..., T]:
    def inside(*args: Any) -> T:
        start_time = time.time()
        return_value = func(*args)
        end_time = time.time()
        duration = end_time - start_time

        print(f"\nDone in {round(duration, 2)}s.")

        return return_value
    return inside

def timed_inline(func: Callable[..., T]) -> Callable[..., T]:
    def inside(*args: Any, verbose: bool=False, label:str, **kwargs: Any):
        print(label)

        start_time = time.time()
        return_value = func(*args, verbose=verbose, **kwargs)
        end_time = time.time()
        duration = end_time - start_time

        if verbose:
            print(f"{label} finished in {round(duration, 2)}s")
        else:
            move_cursor_up_one_line = "\x1b[1A"
            print(f'{move_cursor_up_one_line}{label} {dim(f"{round(duration, 2)}s")}')

        return return_value
    return inside



# type: ignore

# This file has been modified from the RustPython code, which was most likely created mostly from the CPython code
# The license to CPython can be found here: kybra/licenses/CPYTHON_LICENSE
# The license to RustPython can be found here: kybra/licenses/RUST_PYTHON_LICENSE

import sys

class Optional():
    def __class_getitem__(cls, x):
        return cls

class Callable():
    def __class_getitem__(cls, x):
        return True

class Generator():
    def __class_getitem__(cls, x):
        return cls

class Generic():
    def __class_getitem__(cls, x):
        return cls

class Union():
    def __class_getitem__(cls, x):
        return cls

def TypeAlias(self, parameters):
    raise TypeError(f"{self} is not subscriptable")

def TypeVar(x):
    pass

def ParamSpec(x):
    pass

Type = type

def Any(self, parameters):
    raise TypeError(f"{self} is not subscriptable")

def NoReturn(self, parameters):
    raise TypeError(f"{self} is not subscriptable")

def _type_check(arg, msg, is_argument=True, module=None, *, allow_special_forms=False):
    return True

class _TypedDictMeta(type):
    def __new__(cls, name, bases, ns, total=True):
        for base in bases:
            if type(base) is not _TypedDictMeta:
                raise TypeError('cannot inherit from both a TypedDict type '
                                'and a non-TypedDict base class')
        tp_dict = type.__new__(_TypedDictMeta, name, (dict,), ns)

        annotations = {}
        own_annotations = ns.get('__annotations__', {})
        own_annotation_keys = set(own_annotations.keys())
        msg = "TypedDict('Name', {f0: t0, f1: t1, ...}); each t must be a type"
        own_annotations = {
            n: _type_check(tp, msg, module=tp_dict.__module__)
            for n, tp in own_annotations.items()
        }
        required_keys = set()
        optional_keys = set()

        for base in bases:
            annotations.update(base.__dict__.get('__annotations__', {}))
            required_keys.update(base.__dict__.get('__required_keys__', ()))
            optional_keys.update(base.__dict__.get('__optional_keys__', ()))

        annotations.update(own_annotations)
        if total:
            required_keys.update(own_annotation_keys)
        else:
            optional_keys.update(own_annotation_keys)

        tp_dict.__annotations__ = annotations
        tp_dict.__required_keys__ = frozenset(required_keys)
        tp_dict.__optional_keys__ = frozenset(optional_keys)
        if not hasattr(tp_dict, '__total__'):
            tp_dict.__total__ = total
        return tp_dict

    __call__ = dict  # static method

    def __subclasscheck__(cls, other):
        # Typed dicts are only for static structural subtyping.
        raise TypeError('TypedDict does not support instance and class checks')

    __instancecheck__ = __subclasscheck__

def TypedDict(typename, fields=None, /, *, total=True, **kwargs):
    """A simple typed namespace. At runtime it is equivalent to a plain dict.

    TypedDict creates a dictionary type that expects all of its
    instances to have a certain set of keys, where each key is
    associated with a value of a consistent type. This expectation
    is not checked at runtime but is only enforced by type checkers.
    Usage::

        class Point2D(TypedDict):
            x: int
            y: int
            label: str

        a: Point2D = {'x': 1, 'y': 2, 'label': 'good'}  # OK
        b: Point2D = {'z': 3, 'label': 'bad'}           # Fails type check

        assert Point2D(x=1, y=2, label='first') == dict(x=1, y=2, label='first')

    The type info can be accessed via the Point2D.__annotations__ dict, and
    the Point2D.__required_keys__ and Point2D.__optional_keys__ frozensets.
    TypedDict supports two additional equivalent forms::

        Point2D = TypedDict('Point2D', x=int, y=int, label=str)
        Point2D = TypedDict('Point2D', {'x': int, 'y': int, 'label': str})

    By default, all keys must be present in a TypedDict. It is possible
    to override this by specifying totality.
    Usage::

        class point2D(TypedDict, total=False):
            x: int
            y: int

    This means that a point2D TypedDict can have any of the keys omitted.A type
    checker is only expected to support a literal False or True as the value of
    the total argument. True is the default, and makes all items defined in the
    class body be required.

    The class syntax is only supported in Python 3.6+, while two other
    syntax forms work for Python 2.7 and 3.2+
    """
    if fields is None:
        fields = kwargs
    elif kwargs:
        raise TypeError("TypedDict takes either a dict or keyword arguments,"
                        " but not both")

    ns = {'__annotations__': dict(fields)}
    try:
        # Setting correct module is necessary to make typed dict classes pickleable.
        ns['__module__'] = sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return _TypedDictMeta(typename, (), ns, total=total)

_TypedDict = type.__new__(_TypedDictMeta, 'TypedDict', (), {})
TypedDict.__mro_entries__ = lambda bases: (_TypedDict,)



from kybra import blob, Canister, Func, method, nat32, nat64, opt, Principal, Query, Record, Variant
from typing import TypeAlias

# Amount of tokens, measured in 10^-8 of a token.
class Tokens(Record):
    e8s: nat64

# Number of nanoseconds from the UNIX epoch in UTC timezone.
class TimeStamp(Record):
    timestamp_nanos: nat64

# AccountIdentifier is a 32-byte array.
# The first 4 bytes is big-endian encoding of a CRC32 checksum of the last 28 bytes.
AccountIdentifier = blob

# Subaccount is an arbitrary 32-byte byte array.
# Ledger uses subaccounts to compute the source address, which enables one
# principal to control multiple ledger accounts.
SubAccount = blob

# Sequence number of a block produced by the ledger.
BlockIndex = nat64

# An arbitrary number associated with a transaction.
# The caller can set it in a `transfer` call as a correlation identifier.
Memo = nat64

# Arguments for the `transfer` call.
class TransferArgs(Record):
    # Transaction memo.
    # See comments for the `Memo` type.
    memo: Memo
    # The amount that the caller wants to transfer to the destination address.
    amount: Tokens
    # The amount that the caller pays for the transaction.
    # Must be 10000 e8s.
    fee: Tokens
    # The subaccount from which the caller wants to transfer funds.
    # If null, the ledger uses the default (all zeros) subaccount to compute the source address.
    # See comments for the `SubAccount` type.
    from_subaccount: opt[SubAccount]
    # The destination account.
    # If the transfer is successful, the balance of this address increases by `amount`.
    to: AccountIdentifier
    # The point in time when the caller created this request.
    # If null, the ledger uses current IC time as the timestamp.
    created_at_time: opt[TimeStamp]

class TransferError_BadFee(Record):
    expected_fee: Tokens

class TransferError_InsufficientFunds(Record):
    balance: Tokens

class TransferError_TxTooOld(Record):
    allowed_window_nanos: nat64

class TransferError_TxDuplicate(Record):
    duplicate_of: BlockIndex

class TransferError(Variant, total=False):
    # The fee that the caller specified in the transfer request was not the one that ledger expects.
    # The caller can change the transfer fee to the `expected_fee` and retry the request.
    BadFee: TransferError_BadFee
    # The account specified by the caller doesn't have enough funds.
    InsufficientFunds: TransferError_InsufficientFunds
    # The request is too old.
    # The ledger only accepts requests created within 24 hours window.
    # This is a non-recoverable error.
    TxTooOld: TransferError_TxTooOld
    # The caller specified `created_at_time` that is too far in future.
    # The caller can retry the request later.
    TxCreatedInFuture: None
    # The ledger has already executed the request.
    # `duplicate_of` field is equal to the index of the block containing the original transaction.
    TxDuplicate: TransferError_TxDuplicate

class TransferResult(Variant, total=False):
    Ok: nat64
    Err: TransferError

# Arguments for the `account_balance` call.
class AccountBalanceArgs(Record):
    account: AccountIdentifier

class TransferFeeArg(Record): ...

class TransferFee(Record):
    # The fee to pay to perform a transfer
    transfer_fee: Tokens

class GetBlocksArgs(Record):
    # The index of the first block to fetch.
    start: BlockIndex
    # Max number of blocks to fetch.
    length: nat64

class Operation_Mint(Record):
    to: AccountIdentifier
    amount: Tokens

class Operation_Burn(Record):
    from_: AccountIdentifier
    amount: Tokens

class Operation_Transfer(Record):
    from_: AccountIdentifier
    to: AccountIdentifier
    amount: Tokens
    fee: Tokens

class Operation(Variant, total=False):
    Mint: Operation_Mint
    Burn: Operation_Burn
    Transfer: Operation_Transfer

class Transaction(Record):
    memo: Memo
    operation: opt[Operation]
    created_at_time: TimeStamp

class Block(Record):
    parent_hash: opt[blob]
    transaction: Transaction
    timestamp: TimeStamp

# A prefix of the block range specified in the [GetBlocksArgs] request.
class BlockRange(Record):
    # A prefix of the requested block range.
    # The index of the first block is equal to [GetBlocksArgs.from].
    #
    # Note that the number of blocks might be less than the requested
    # [GetBlocksArgs.len] for various reasons, for example:
    #
    # 1. The query might have hit the replica with an outdated state
    #    that doesn't have the full block range yet.
    # 2. The requested range is too large to fit into a single reply.
    #
    # NOTE: the list of blocks can be empty if:
    # 1. [GetBlocksArgs.len] was zero.
    # 2. [GetBlocksArgs.from] was larger than the last block known to the canister.
    blocks: list[Block]

class QueryArchiveError_BadFirstBlockIndex(Record):
    requested_index: BlockIndex
    first_valid_index: BlockIndex

class QueryArchiveError_Other(Record):
    error_code: nat64
    error_message: str

# An error indicating that the arguments passed to [QueryArchiveFn] were invalid.
class QueryArchiveError(Variant, total=False):
    # [GetBlocksArgs.from] argument was smaller than the first block
    # served by the canister that received the request.
    BadFirstBlockIndex: QueryArchiveError_BadFirstBlockIndex
    # Reserved for future use.
    Other: QueryArchiveError_Other

class QueryArchiveResult(Variant, total=False):
    # Successfully fetched zero or more blocks.
    Ok: BlockRange
    # The [GetBlocksArgs] request was invalid.
    Err: QueryArchiveError

# A function that is used for fetching archived ledger blocks.
QueryArchiveFn: TypeAlias = Func(Query[[GetBlocksArgs], QueryArchiveResult])

class QueryBlocksResponse_archived_blocks(Record):
    # The index of the first archived block that can be fetched using the callback.
    start: BlockIndex

    # The number of blocks that can be fetch using the callback.
    length: nat64

    # The function that should be called to fetch the archived blocks.
    # The range of the blocks accessible using this function is given by [from]
    # and [len] fields above.
    callback: QueryArchiveFn

# The result of a "query_blocks" call.
#
# The structure of the result is somewhat complicated because the main ledger canister might
# not have all the blocks that the caller requested: One or more "archive" canisters might
# store some of the requested blocks.
#
# Note: as of Q4 2021 when this interface is authored, the IC doesn't support making nested
# query calls within a query call.
class QueryBlocksResponse(Record):
    # The total number of blocks in the chain.
    # If the chain length is positive, the index of the last block is `chain_len - 1`.
    chain_length: nat64

    # System certificate for the hash of the latest block in the chain.
    # Only present if `query_blocks` is called in a non-replicated query context.
    certificate: opt[blob]

    # List of blocks that were available in the ledger when it processed the call.
    #
    # The blocks form a contiguous range, with the first block having index
    # [first_block_index] (see below), and the last block having index
    # [first_block_index] + len(blocks) - 1.
    #
    # The block range can be an arbitrary sub-range of the originally requested range.
    blocks: list[Block]

    # The index of the first block in "blocks".
    # If the blocks vector is empty, the exact value of this field is not specified.
    first_block_index: BlockIndex

    # Encoding of instructions for fetching archived blocks whose indices fall into the
    # requested range.
    #
    # For each entry `e` in [archived_blocks], `[e.from, e.from + len)` is a sub-range
    # of the originally requested block range.
    archived_blocks: list[QueryBlocksResponse_archived_blocks]

class Archive(Record):
    canister_id: Principal

class Archives(Record):
    archives: list[Archive]

class SymbolResult(Record):
    symbol: str

class NameResult(Record):
    name: str

class DecimalsResult(Record):
    decimals: nat32

Address = str

class Ledger(Canister):
    # Transfers tokens from a subaccount of the caller to the destination address.
    # The source address is computed from the principal of the caller and the specified subaccount.
    # When successful, returns the index of the block containing the transaction.
    @method
    def transfer(self, transfer_args: TransferArgs) -> TransferResult: ...

    # Returns the amount of Tokens on the specified account.
    @method
    def account_balance(self, account_balance_args: AccountBalanceArgs) -> Tokens: ...

    # Returns the current transfer_fee.
    @method
    def transfer_fee(self, transfer_fee_arg: TransferFeeArg) -> TransferFee: ...

    # Queries blocks in the specified range.
    @method
    def query_blocks(self, get_blocks_args: GetBlocksArgs) -> QueryBlocksResponse: ...

    # Returns token symbol.
    @method
    def symbol(self) -> SymbolResult: ...

    # Returns token name.
    @method
    def name(self) -> NameResult: ...

    # Returns token decimals.
    @method
    def decimals(self) -> DecimalsResult: ...

    # Returns the existing archive canisters information.
    @method
    def archives(self) -> Archives: ...




# The Principal code is taken from https://github.com/rocklabs-io/ic-py/blob/main/ic/principal.py

# MIT License

# Copyright (c) 2021 Rocklabs

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# principal type: https://github.com/dfinity/ic-types/blob/main/src/principal.rs

import zlib
import math
# import base64 # TODO we probably need to remove this
import binascii
import hashlib
from typing import Union
# from enum import Enum # TODO we probably need to remove this

CRC_LENGTH_IN_BYTES = 4
HASH_LENGTH_IN_BYTES = 28
MAX_LENGTH_IN_BYTES = 29

# class PrincipalClass(Enum):
#     OpaqueId = 1
#     SelfAuthenticating = 2
#     DerivedId = 3
#     Anonymous = 4
#     # Unassigned

class Principal:
    def __init__(self, bytes: bytes = b''):
        self._len = len(bytes)
        self._bytes = bytes
        self.hex = str(self._bytes.hex()).upper()
        self._isPrincipal = True

    @staticmethod
    def management_canister():
        return Principal()

    @staticmethod
    def self_authenticating(pubkey: Union[str, bytes]):
        if isinstance(pubkey, str):
            pubkey = bytes.fromhex(pubkey)
        hash_ = hashlib.sha224(pubkey).digest()
        hash_ += bytes([2])
        return Principal(bytes = hash_)

    @staticmethod
    def anonymous():
        return Principal(bytes = b'\x04')

    @property
    def len(self):
        return self._len

    @property
    def bytes(self):
        return self._bytes

    @property
    def isPrincipal(self):
        return self._isPrincipal

    @staticmethod
    def from_str(s: str):
        s1 = s.replace('-', '')
        pad_len = math.ceil(len(s1) / 8) * 8 - len(s1)
        b = b32decode(s1.upper().encode() + b'=' * pad_len)
        if len(b) < CRC_LENGTH_IN_BYTES:
            raise Exception("principal length error")
        p = Principal(bytes = b[CRC_LENGTH_IN_BYTES:])
        if not p.to_str() == s:
            raise Exception("principal format error")
        return p

    @staticmethod
    def from_hex(s: str):
        return Principal(bytes.fromhex(s.lower()))

    def to_str(self):
        checksum = zlib.crc32(self._bytes) & 0xFFFFFFFF
        b = b''
        b += checksum.to_bytes(CRC_LENGTH_IN_BYTES, byteorder='big')
        b += self.bytes
        s = b32encode(b).decode('utf-8').lower().replace('=', '')
        ret = ''
        while len(s) > 5:
            ret += s[:5]
            ret += '-'
            s = s[5:]
        ret += s
        return ret

    def to_account_id(self, sub_account: int = 0):
        return AccountIdentifier.new(self, sub_account)

    def __repr__(self):
        return "Principal(" + self.to_str() + ")"

    def __str__(self):
        return self.to_str()

class AccountIdentifier:
    def __init__(self, hash: bytes) -> None:
        assert len(hash) == 32
        self._hash = hash

    def to_str(self):
        return '0x' + self._hash.hex()

    def __repr__(self):
        return "Account(" + self.to_str() + ")"

    def __str__(self):
        return self.to_str()

    @property
    def bytes(self) -> bytes:
        return self._hash

    @staticmethod
    def new(principal: Principal, sub_account: int = 0):
        sha224 = hashlib.sha224()
        sha224.update(b'\x0Aaccount-id')
        sha224.update(principal.bytes)

        sub_account = sub_account.to_bytes(32, byteorder='big')
        sha224.update(sub_account)
        hash = sha224.digest()
        checksum = zlib.crc32(hash) & 0xFFFFFFFF
        account = checksum.to_bytes(CRC_LENGTH_IN_BYTES, byteorder='big') + hash
        return AccountIdentifier(account)

# Everything below taken from RustPython/Lib/base64.py

# These functions have been modified from the RustPython code, which was most likely created mostly from the CPython code
# The license to CPython can be found here: kybra/licenses/CPYTHON_LICENSE
# The license to RustPython can be found here: kybra/licenses/RUST_PYTHON_LICENSE

# Base32 encoding/decoding must be done in Python
_b32alphabet = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
_b32tab2 = None
_b32rev = None

bytes_types = (bytes, bytearray)  # Types acceptable as binary data

def b32encode(s):
    """Encode the bytes-like object s using Base32 and return a bytes object.
    """
    global _b32tab2
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    if _b32tab2 is None:
        b32tab = [bytes((i,)) for i in _b32alphabet]
        _b32tab2 = [a + b for a in b32tab for b in b32tab]
        b32tab = None

    if not isinstance(s, bytes_types):
        s = memoryview(s).tobytes()
    leftover = len(s) % 5
    # Pad the last quantum with zero bits if necessary
    if leftover:
        s = s + b'\0' * (5 - leftover)  # Don't use += !
    encoded = bytearray()
    from_bytes = int.from_bytes
    b32tab2 = _b32tab2
    for i in range(0, len(s), 5):
        c = from_bytes(s[i: i + 5], 'big')
        encoded += (b32tab2[c >> 30] +           # bits 1 - 10
                    b32tab2[(c >> 20) & 0x3ff] + # bits 11 - 20
                    b32tab2[(c >> 10) & 0x3ff] + # bits 21 - 30
                    b32tab2[c & 0x3ff]           # bits 31 - 40
                   )
    # Adjust for any leftover partial quanta
    if leftover == 1:
        encoded[-6:] = b'======'
    elif leftover == 2:
        encoded[-4:] = b'===='
    elif leftover == 3:
        encoded[-3:] = b'==='
    elif leftover == 4:
        encoded[-1:] = b'='
    return bytes(encoded)

def b32decode(s, casefold=False, map01=None):
    """Decode the Base32 encoded bytes-like object or ASCII string s.

    Optional casefold is a flag specifying whether a lowercase alphabet is
    acceptable as input.  For security purposes, the default is False.

    RFC 3548 allows for optional mapping of the digit 0 (zero) to the
    letter O (oh), and for optional mapping of the digit 1 (one) to
    either the letter I (eye) or letter L (el).  The optional argument
    map01 when not None, specifies which letter the digit 1 should be
    mapped to (when map01 is not None, the digit 0 is always mapped to
    the letter O).  For security purposes the default is None, so that
    0 and 1 are not allowed in the input.

    The result is returned as a bytes object.  A binascii.Error is raised if
    the input is incorrectly padded or if there are non-alphabet
    characters present in the input.
    """
    global _b32rev
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    if _b32rev is None:
        _b32rev = {v: k for k, v in enumerate(_b32alphabet)}
    s = _bytes_from_decode_data(s)
    if len(s) % 8:
        raise binascii.Error('Incorrect padding')
    # Handle section 2.4 zero and one mapping.  The flag map01 will be either
    # False, or the character to map the digit 1 (one) to.  It should be
    # either L (el) or I (eye).
    if map01 is not None:
        map01 = _bytes_from_decode_data(map01)
        assert len(map01) == 1, repr(map01)
        s = s.translate(bytes.maketrans(b'01', b'O' + map01))
    if casefold:
        s = s.upper()
    # Strip off pad characters from the right.  We need to count the pad
    # characters because this will tell us how many null bytes to remove from
    # the end of the decoded string.
    l = len(s)
    s = s.rstrip(b'=')
    padchars = l - len(s)
    # Now decode the full quanta
    decoded = bytearray()
    b32rev = _b32rev
    for i in range(0, len(s), 8):
        quanta = s[i: i + 8]
        acc = 0
        try:
            for c in quanta:
                acc = (acc << 5) + b32rev[c]
        except KeyError:
            raise binascii.Error('Non-base32 digit found') from None
        decoded += acc.to_bytes(5, 'big')
    # Process the last, partial quanta
    if l % 8 or padchars not in {0, 1, 3, 4, 6}:
        raise binascii.Error('Incorrect padding')
    if padchars and decoded:
        acc <<= 5 * padchars
        last = acc.to_bytes(5, 'big')
        leftover = (43 - 5 * padchars) // 8  # 1: 4, 3: 3, 4: 2, 6: 1
        decoded[-5:] = last[:leftover]
    return bytes(decoded)

def _bytes_from_decode_data(s):
    if isinstance(s, str):
        try:
            return s.encode('ascii')
        except UnicodeEncodeError:
            raise ValueError('string argument should contain only ASCII characters')
    if isinstance(s, bytes_types):
        return s
    try:
        return memoryview(s).tobytes()
    except TypeError:
        raise TypeError("argument should be a bytes-like object or ASCII "
                        "string, not %r" % s.__class__.__name__) from None



from typing import TypedDict

class Args(TypedDict):
    empty: bool
    flags: "Args_flags"
    canister_name: str
    entry_point: str
    did_path: str

class Args_flags(TypedDict):
    verbose: bool

class Paths(TypedDict):
    py_entry_file: str
    py_entry_module_name: str
    canister: str
    python_source: str
    py_file_names_file: str
    did: str
    compiler: str
    lib: str
    generated_did: str
    target: str
    wasm: str
    gzipped_wasm: str
    custom_modules: str
